
/** @file mlx_dev_emu.h
 * Mellanox Device Emulation SDK.
 *
 * ## Overview
 *
 * This library allows emulating a PCI device using Mellanox BlueField-based SmartNIC.
 * It is intended for use by an application running on the SmartNIC's internal processor,
 * and provides the needed functionality for that application to act as a PCI
 * device toward the host where the SmartNIC is located.
 * This application is called the "emulation software" in this API doc.
 *
 * The supplied functionality includes interacting with emulated registers presented to
 * the host, getting notified of doorbells or other actions initiated by the
 * host, DMAing data between host memory and internal processor memory and
 * signaling the host with interrupt.
 *
 * The API is made of several layers. A basic layer that allows emulating
 * generic devices (but not all possible devices!), and layers that are more
 * efficient for specific types of devices, such as NVMe emulation. These APIs
 * use specific offloads provided by the SmartNIC to allow such efficiency.
 * 
 * ## Devices Model
 *
 * SmartNIC is preconfigured (using Mellanox nvconfig tool) to present emulated
 * device PFs and VFs. For each emulated PF presented to the host, an RDMA PF
 * is presented to the SmartNIC internal processor. The emulated PF and all its
 * VFs are managed and serviced using that RDMA PF inside the SmartNIC.
 *
 * ## Working with emulated devices
 *
 * An emulated device has a BAR with registers space. The emulation software
 * using the SDK can communicate with the host using those registers.
 * The registers act as a memory: both host and emulation software reads/writes
 * to them act as reads/writes to shared memory space.
 *
 * It is expected that the device defines an efficient way for communications
 * with the host using queues. Each such queue is made of a piece of host
 * memory that acts like a queue and a doorbell. Host should be writing
 * elements to the queue, and then write the last producer index to the
 * doorbell. The SDK will facilitate transferring those queue posts to the
 * emulation software.

 * ## Threads Model
 *
 * A worker context is used to describe a single thread. This thread should be
 * responsible for a group of emulated queues created for a specific emulated
 * device.
 * The thread should periodically call the progression function in this SDK
 * so that the different callbacks can be called based on available events from
 * the queues bound to the thread.
 * For events driven model, a file descriptor can be used to poll/epoll/select
 * on, which will wake up the thread when a new event is waiting. The thread
 * is expected to run the progression function when waking up.
 *
 * ## Types of emulation
 *
 * The most basic emulation type delivers the doorbells that the host is
 * writing to the emulation software. The emulation software is then expected
 * to use the DMA capabilities in the SDK to bring the queue elements, parse
 * them, act on them, potentially DMA more data from/to host, and eventually
 * DMA a completion item to the emulated completion queue and signal an
 * interrupt. In this mode, the completion queue is another emulated queue
 * that repors doorbells when the host consumes a completion item.
 *
 * More advanced modes of emulation adds knowledge on the queues and the
 * elements posted to them, and make it more efficient for the emulation
 * software.
 *
 * ## Flow
 *
 * The general flow of calls to emulate a device is:
 * 1. Call mlx_emu_get_version() to get library version and capabilities.
 * 2. Call mlx_emu_dev_query() to get device emulation capabilities and status of a specific RDMA NIC function.
 * 3. Call mlx_emu_dev_open() to "open" a specific emulated device function and begin working with it.
 * 4. Call mlx_emu_worker_create() to create a worker thread context.
 * 5. Optionally call mlx_emu_worker_fd() to get a file descriptor to sleep on (not yet supported). Otherwise work in polling mode.
 * 6. Call mlx_emu_worker_progress() periodically or when file descriptor is ready for read.
 * 7. Call mlx_emu_bar_read() and mlx_emu_bar_write() to communicate with the host using device registers.
 * 8. Call mlx_emu_q_create() to create an emulated queue. SDK will now track this queue (doorbell rings or more).
 * 9. Handle the callback whenever a new work is posted by the host to the emulated queue.
 * 10. Use mlx_emu_q_dma_read_bcopy() and mlx_emu_q_dma_write_bcopy() to DMA queue elements and other data between
 *     emulation software and host memory.
 * 11. Call mlx_emu_q_interrupt() to signal an interrupt to the host. i.e, after DMA of a completion element.
 *
 */

#ifndef MLX_DEV_EMU_H
#define MLX_DEV_EMU_H

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

/**
 * Status return codes.
 */
enum {
    MLX_EMU_OK  =  0,
    MLX_EMU_ERR = -1,
    MLX_EMU_NOTSUPPORTED = -2,
};


#define MLX_DEV_EMU_LOCK_FD_PREFIX "/var/lock"

#define MLX_DEV_EMU_VER_MAJOR   0
#define MLX_DEV_EMU_VER_MINOR   0
#define MLX_DEV_EMU_VER_SUB     1

/**
 * Device Emulation Attributes.
 *
 * Describes basic attributes and capabilities of the device emulation.
 */
typedef struct {
    struct {
        uint16_t major;
        uint16_t minor;
        uint16_t sub;
    } v;		/**< version number */
    uint64_t caps;      /**< capabilitis mask, not yet implemented */
    size_t   max_bcopy; /**< max size that can be transfered via bcopy read/write */
} mlx_emu_ver_t;

/**
 * Get library version and capabilities.
 * @param[out] attr Pointer to API attributes
 */
int mlx_emu_get_version(mlx_emu_ver_t *attr);

typedef struct mlx_emu_dev     mlx_emu_dev_t; /**< Emulated device opaque context */
typedef struct mlx_emu_worker  mlx_emu_worker_t; /**< Worker opaque context */
typedef struct mlx_emu_q       mlx_emu_q_t; /**< Emulated queue opaque context */

/**
 * Callback function indicating registers space of emulated device was written by host.
 *
 * Should be supplied when calling mlx_emu_dev_open().
 * Called when host has written to the bar registers space, which is on BAR 0
 * from the beginning and up to mlx_emu_dev_attr_t.bar_reg_size supplied to mlx_emu_dev_open().
 * Since registers are per device, this callback is always called from the
 * context for the first worker.
 */
typedef void (*mlx_emu_bar_write_cb_t)(mlx_emu_worker_t *worker);

/**
 * Callback function indicating that FLR (function level reset) was
 * triggered.
 *
 * @param[in] status will be MLX_EMU_FLR_START when FLR is detected and
 *            MLX_EMU_FLR_COMPLETED when library is ready to be used
 */
enum {
    MLX_EMU_FLR_START,
    MLX_EMU_FLR_COMPLETED,
};

typedef void (*mlx_emu_flr_cb_t)(mlx_emu_worker_t *worker, int status);

/**
 * A doorbell callback.
 * Called when host has written to the doorbell address of the emulated
 * queue.
 * Called from the context of the worker that the emulated queue belongs to.
 *
 * @param[in] q the emulated queue that is the source of this doorbell.
 * @param[in] db_addr the doorbell address that was written to.
 * @param[in] db_val the doorbell value (producer index) that was written.
 */
typedef void (*mlx_emu_q_db_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, uint32_t db_val);


/**
 * NVME command to NVMF command capsule conversion callback
 */
typedef void (*mlx_emu_q_nvmf_cmd_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, void *cmd, uint32_t len);

/**
 * NVME command callback
 */
typedef void (*mlx_emu_q_nvme_cmd_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, void *cmd, uint32_t len);

enum {
    MLX_EMU_ASYNC_EV_QUEUE_ERROR    /**< A generic queue error */
};

/**
 * An asynchronous event callback 
 */
typedef void (*mlx_emu_q_async_ev_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, int event);

/**
 * Queue emulation types.
 */
enum {
    MLX_EMU_CB_TYPE_DB,	                /**< Only pass doorbell notifications */
    MLX_EMU_CB_TYPE_NVME,               /**< Pass raw NVME SQE capsule */
    MLX_EMU_CB_TYPE_NVME2NVMF,          /**< Convert NVME command to NVMF command capsule */
};

/**
 * A basic structure for queue emulation callbacks.
 *
 * Should not be used directly. Holds nothing but the type of emulation.
 * Use the mlx_emu_xxx_mode_db_t structures instead.
 */
typedef struct {
   uint8_t emu_cb_type;	/**< Emulation type, from the MLX_EMU_CB_TPYE_xxx enumbs */
} mlx_emu_cb_base_t; 

/**
 * A queue emulation of type doorbell callbacks structure
 */
typedef struct {
    mlx_emu_cb_base_t   super;	/**< the basic structure which holds the emulation type */
    mlx_emu_q_db_cb_t   doorbell_write_cb;	/**< a callback for delivering a doorbell notification */
} mlx_emu_doorbell_mode_cb_t;

/**
 * NVME to NVMF  submission queue emulation
 */
typedef struct {
    mlx_emu_cb_base_t        super;         
    mlx_emu_q_nvmf_cmd_cb_t  nvmf_command_cb;  /**< callback for delivering a NVME command converted 
                                                    to a NVMF command capsule */
} mlx_emu_nvmf_mode_cb_t;

/**
 * NVME submission queue emulation
 */
typedef struct {
    mlx_emu_cb_base_t        super;         
    mlx_emu_q_nvme_cmd_cb_t  nvme_command_cb;  /**< callback for delivering a NVME command */ 
} mlx_emu_nvme_mode_cb_t;

/**
 * Emulated device attributes.
 *
 * Supplied when opening a speicifc emulated device function, and includes
 * parameters. 
 */
typedef struct {
    uint16_t function;       /**< function number of the emulated device to open */
    uint16_t bar_reg_size;   /**< actual size of the registers space in the BAR */
    uint16_t db_base;        /**< offset from the beginnig of BAR 0 where doorbells start */
    uint64_t caps;	     /**< required capabilities */
    /* lock policy ? */
    mlx_emu_bar_write_cb_t    bar_write_cb; /**< a callback function to be called whenever the registers space has been written by the host */
} mlx_emu_dev_attr_t;

/**
 * Device Emulation capabilities and attributes of specific RDMA device.
 *
 */
typedef struct {
    uint16_t max_functions;     /**< number of emulated functions that are presented to the host */
    uint32_t max_namespaces;    /**< max number of namespaces support per SQ */
    uint16_t max_bar_reg_size;  /**< max size of the bar register area */
    uint32_t max_emulated_cq;   /**< max number of CQs that can be emulated */
    uint32_t max_emulated_sq;   /**< max number of SQs that can be emulated */
    uint16_t max_workers;       /**< max number of workers */
    uint64_t caps;              /**< emulation capabilities  */
} mlx_emu_dev_query_attr_t;

/**
 * Device Emulation queue types
 */
enum mlx_emu_q_flags {
    MLX_EMU_Q_NONE        = 0,      /**< A default queue */
    MLX_EMU_Q_CQ          = (1<<0), /**< The queue represents a completion queue */
    MLX_EMU_Q_SQ          = (1<<1), /**< The queue is a submission queue */
    MLX_EMU_Q_EXTERNAL_QP = (1<<2)  /**< The queue uses an external QP to transfer data */
};

typedef struct {
	uint32_t src_nsid;
	uint32_t dst_nsid;
	uint16_t metadata_size;
	uint8_t lba_size;
} nvme_namespace_format_t;

/**
 * Attributes of an emulated queue.
 *
 * An emulated queue refers to a memory region in host memory acting as a queue
 * with a doorbell address where the host reports producer index.
 * When creating an emulated queue different information about the queue is
 * needed  depending on the mode of emulation requested. A different
 * callback is provided to deliver the notifications of submittions to the
 * queue.
 */
typedef struct {
    uint32_t                 db_offset;  /**< doorbell offset relative to the device 
                                           doorbell base. Always needed */
    uint16_t                 max_length; /**< Max queue length. Always needed. */
    uint32_t                 flags;       /**< See @ref enum mlx_emu_q_flags. Always needed */
    int                      msi_x;	     /**< msix interrupt to generate when requested.
                                           Should be set when this queue is a 
                                           a completion queue */
    uint64_t				 base_addr;  /**< sqe base physical address */
    uint8_t					 log_entry_size; /**< log(base 2) of the sqe entry size. */
    uint8_t					 log_page_size; /**< Log(base 2) of the page size host is
     	 	 	 	 	 	 	 	 	     working with the NVMe device in granularity of 4KB.
     	 	 	 	 	 	 	 	 	      i.e. page size in bytes is 2^(12+log_nvme_page_size). */
    uint8_t 				 max_transcation_size; /**< Device will verify that each SQE does not exceeds
    												this size in granularity of 2^log_nvme_page_size. */
    uint16_t 				 num_of_namespaces; /**< Number of namespaces entries in list. */
    nvme_namespace_format_t *nvme_namespace_l; /**< Namespace entries List */
    uint32_t                 ext_qp_num; /**< if non zero and MLX_EMU_Q_EXTERNAL_QP flag is set
                                           SDK will use the qp to send all 
                                           emulation related data such as doorbells, NVMoF
                                           command capsules etc. The caller is responsible for
                                           the qp lifecyle management. E.x. connecting to the
                                           remote qp, handling errors, receiving data etc...
                                           The callbacks normally used to deliver commands 
                                           (such as mlx_emu_q_nvmf_cmd_cb_t) will never be called.
                                           When external QP is used, there is no sense in having more
                                           than one worker that will handle async events and registers 
                                           events, so all queues can be created using the same worker context. 
                                          */
    mlx_emu_q_t             *cq;         /**< If the queue is a submission queue it should point to
                                              a valid completion queue */ 
    mlx_emu_cb_base_t       *cb;         /**< mode of emulation and related callbacks
                                           and info */
    void                    *ucontext;   /**< user context */
    mlx_emu_q_async_ev_cb_t *async_ev_cb; /**< callback which is called when async event or
                                          error is detected on the queue */
    uint8_t                 cq_period;
    uint8_t                 cq_max_count;
} mlx_emu_q_attr_t;

/**
 * Query device emulation capabilities and status of a specific RDMA device.
 *
 * Emulation software should run this query per every RDMA device available
 * to the SmartNIC cpu prior to initializing the device. Those RDMA devices
 * that return successful value of MLX_EMU_OK are in charge of emulated
 * devices, and the emulation software should work with them to service the
 * emulation. The returned dev_caps indicates how many such emulated functions
 * this RDMA device is in charge of and the emulation software should service.
 *
 * Other RDMA devices either do not present any emulated devices or don't
 * support emulation at all.
 *
 * @param[in] dev_name The name of the RDMA device to query, such as "mlx5_2".
 * @param[out] dev_caps The emulation capabilities and status avaiable by the RDMA device.
 * @retval MLX_EMU_OK Device emulation is supported by this RDMA device
 * @retval MLX_EMU_NOTSUPPORTED Device emulation is not supported by this RDMA device
 * @retval MLX_ERR Error while querying the RDMA device
 */
int mlx_emu_dev_query(const char *dev_name, mlx_emu_dev_query_attr_t *dev_caps);

/**
 * Open a specific emulated device function to begin working with it.
 *
 * Emulation software should call this for each of the emulated functions
 * in order to begin working with them.
 *
 * This call will check the supplied attributes and then allocate a new context 
 * indentifying the emulated device function for the rest of this API.
 *
 * @param[in] dev_name RDMA device name, such as "mlx5_2".
 * @param[in] arrt Parameters for the emulated device.
 * @param[out] emu_dev_context_p Pointer to a newly allocated device context, if successful.
 *
 * @retval MLX_EMU_OK Sucess, device context allocated and returned via emu_dev_context_p. 
 * @retval MLX_EMU_ERR Error creating or initializing the device context.
 * @retval MLX_ERR Error communicating with RDMA device.
 *
 */
int mlx_emu_dev_open(const char *dev_name, mlx_emu_dev_attr_t *attr, mlx_emu_dev_t **emu_dev_context_p);

/**
 * Close an emulated device context.
 * 
 * Close and deallocate emulated device context, previously opened using mlx_emu_dev_open()
 *
 */
int mlx_emu_dev_close(mlx_emu_dev_t *emu_dev_context);

int mlx_emu_dev_start(mlx_emu_dev_t *dev_ctx);
void mlx_emu_dev_stop(mlx_emu_dev_t *dev_ctx);
int mlx_emu_dev_wait_ready(mlx_emu_dev_t *dev_ctx);

/**
 * Create a worker thread context.
 *
 * A worked context should be created for every thread in the emulation
 * software.
 *
 * @param[in] emu_dev_ctx A pointer to the emulated device context
 * @param[in] context     A pointer to the user supplied context
 * @param[in] num_queues  number of queues in the worker
 * @param[in] worker_p    a pointer to the created context pointer
 */
int mlx_emu_worker_create(mlx_emu_dev_t *emu_dev_ctx, void *context, uint32_t num_queues,
                          mlx_emu_worker_t **worker_p);

/**
 * Get worker context data 
 */
void *mlx_emu_worker_context(mlx_emu_worker_t *worker);

/**
 * Destroy a worker thread context.
 *
 * Deallocate a worker context previously created with mlx_emu_worker_create()
 */
int mlx_emu_worker_destroy(mlx_emu_worker_t *worker);

/**
 * Get a file descriptor that can be used with poll/epoll/select.
 *
 * Not yet supported.
 *
 * The descriptor will be ready for read upon arrival of the new events. When
 * this happens emulation software should call the mlx_emu_worker_progress()
 * to handle the events.
 *
 * @param[in] worker A pointer to worker context, created with mlx_emu_worker_create().
 * @param[out] fd A file descriptor to be used in poll/select.
 */
int mlx_emu_worker_fd(mlx_emu_worker_t *worker, int *fd);

/**
 * Progress a worker and handle pending events.
 *  
 * Emulation software should call this function periodically, or when the FD
 * returned from mlx_emu_worker_fd() becomes available.
 * During this call the SDK will sequentially call the needed callbacks that
 * the emulation software provided in order to handle the available events.
 *
 * If there are events pending, this call will be blocking until all related
 * callbacks returned.
 *
 * @return Number of events that were handled.
 */
int mlx_emu_worker_progress(mlx_emu_worker_t *worker);

/**
 * Create an emulated queue.
 *
 * This call sets up an emulated queue to track. The SDK will notify the
 * emulation software of changes happening in the emulated queue, via the
 * callbacks supplied.
 *
 * Depending on the type of emulation requested in the supplied attributes,
 * different information is needed about the emulated queue, and a different
 * notification callback will be called.
 *
 * Supported queue emulation modes are:
 * 1. Doorbell. In this mode, SDK will notify the emulation software whenever
 *    a doorbell was written by the host to the specified doorbell address. The
 *    notification includes the new doorbell value. The emulation software
 *    callback is then required to fetch the actual queue emelents that were
 *    posted by the host based on this value. In this mode the same call and 
 *    mode should be used for the emulated completion queue.
 *
 * As soon as this call returns the emulated queue tracking begins, and calls
 * to mlx_emu_worker_progress() will dispatch the supplied callback.
 *
 * @param[in] worker The worker context this queue handling will belong to.
 *       Events generated by this queue will issue callbacks only
 *       when this worker context is called with mlx_emu_worker_progress().
 * @param[in] attr Attributes for the emulated queue.
 * @param[out] q_p A queue context representing this queue for further
 *       reference in this API.
 */
int mlx_emu_q_create(mlx_emu_worker_t *worker, mlx_emu_q_attr_t *attr,
                     mlx_emu_q_t **q_p, int q_type);

/**
 * Attributes that can be modified by the @ref mlx_emu_q_modify()
 */
enum mlx_emu_q_modify_mask {
    MLX_EMU_Q_MODIFY_EXTERNAL_QP = 1 /** < Change an external QP number */
};

/**
 * Modify an emulated queue
 *
 * Change queue attributes. When changing external qp number it is up to
 * the caller to gurantee that data loss does not happen. 
 * @param[in] attr_mask a bitwise mask of queue atrributes @ref enum mlx_emu_q_modify_mask
 *            that should be modified. At the moment only 
 *            MLX_EMU_Q_MODIFY_EXTERNAL_QP is supported.
 *
 * @retval MLX_EMU_NOTSUPPORTED if queue can not be modified at all
 * @retval MLX_EMU_ERR if there was some kind of error
 */
int mlx_emu_q_modify(mlx_emu_q_t *q, mlx_emu_q_attr_t *attr, uint32_t attr_mask);

/**
 * Query current queue attributes
 */
int mlx_emu_q_query(mlx_emu_q_t *q, mlx_emu_q_attr_t *attrs);

/**
 * Get a user supplied queue context
 *
 * @retval a context supplied during the queue creation
 */
void *mlx_emu_q_context(mlx_emu_q_t *q);

/**
 * Destroy an emulated queue.
 *
 * Destroy an emulated queue previously created using mlx_emu_q_create().
 */
int mlx_emu_q_destroy(mlx_emu_q_t *q);

/**
 * Send interrupt to the host.
 *
 * Use the given queue to send an interrupt to the host. The MSIX interrupt
 * vector that was configured with the queue upon its creation will be used.
 *
 * @param[in] q The queue that should generate an interrupt
 */
void mlx_emu_q_interrupt(mlx_emu_q_t *q);

/**
 * DMA read from host memory.
 * 
 * Read from host memory address and length and copy it to the supplied local buffer.
 * The read is synchronous, which means the call will block untill read is complete.
 * An internal copy of data happens during the call.
 *
 * @param[in] q the emulated queue to use for the DMA.
 * @param[in] dst_buf a destination local buffer to copy the data into.
 * @param[in] srcaddr the physical address of host memory to read from.
 * @param[in] rmkey the remote key for the RDMA (0 means we'll use global key).
 * @param[in] len the length of data to DMA.
 */
void mlx_emu_q_dma_read_bcopy(mlx_emu_q_t *q, void *dst_buf, uint64_t srcaddr,
                              uint32_t rmkey, size_t len);

/**
 * DMA write to host memory.
 * 
 * Write to host memory address and length the supplied local buffer.
 * The write is synchronous, which means the call will block untill write is complete.
 * An internal copy of data happens during the call.
 *
 * @param[in] q the emulated queue to use for the DMA.
 * @param[in] src_buf a source local buffer to copy the data from.
 * @param[in] dstaddr the physical address of host memory to write to.
 * @param[in] rmkey the remote key for the RDMA (0 means we'll use global key).
 * @param[in] len the length of data to DMA.
 */
void mlx_emu_q_dma_write_bcopy(mlx_emu_q_t *q, void *src_buf, uint64_t dstaddr,
                               uint32_t rmkey, size_t len);

/**
 * DMA send to host memory.
 *
 * Send local buffer to host memory
 *
 * @param[in] q the emulated queue to use for the DMA.
 * @param[in] src_buf a source local buffer to copy the data from.
 * @param[in] len the length of data to DMA.
 */
void mlx_emu_q_send_bcopy(mlx_emu_q_t *q, void *src_buf, size_t len);

/**
 * @attention Not implemented.
 */
void mlx_emu_q_dma_flush(mlx_emu_q_t *q);

/**
 * Read host BAR.
 *
 * When the host writes to BAR registers, they are kept by the device until the
 * emulation software uses this call to fetch the latest values.
 *
 * It is possible to read a single register, a bunch of contiguous registers
 * or even the entire registers space at once.
 *
 * @param[in] dev_ctx the emulated device context.
 * @param[in] buf a local buffer to copy the registers values to.
 * @param[in] bar_addr the address to read from the start of the BAR.
 * @param[in] len the length to read.
 */ 
int mlx_emu_bar_read(mlx_emu_dev_t *dev_ctx, void *buf, uint32_t bar_addr, unsigned len);

/**
 * Write to host BAR.
 *
 * When the host reads BAR registers, it gets a value that was previously written
 * to that register either by the host itself or by the emulation software
 * using this call. 
 *
 * It is possible to write a single register, a bunch of contiguous registers
 * or even the entire registers space at once.
 *
 * @param[in] dev_ctx the emulated device context.
 * @param[in] buf a local buffer to copy the registers values from.
 * @param[in] bar_addr the address to write to from the start of the BAR.
 * @param[in] len the length to write.
 */ 
int mlx_emu_bar_write(mlx_emu_dev_t *dev_ctx, void *buf, uint32_t bar_addr, unsigned len);

void mlx_prp_write_short(mlx_emu_q_t *sq, uint64_t prp1, uint64_t prp2, void *buf, size_t len, unsigned pbits);
void mlx_prp_read_short(mlx_emu_q_t *sq, uint64_t prp1, uint64_t prp2, void *buf, size_t len, unsigned pbits);

#endif
